/* Copyright (C) Ken R. Lunde, 1991-1992.  All rights reserved. */

/* These routines originally came from Ken R. Lunde's JIS.C program, */
/* version 2.1, May 7, 1991.  Many thanks to Ken for his donation.   */

/* The original code was reformatted (to fit the style of the rest   */
/* of the program).  Because EUC is used as the internal representa- */
/* tion, code corresponding to conversions not involving EUC is      */
/* deleted.  Also, the original file-based character I/O machanism   */
/* is changed.                                                       */

/* A modified version of the original documentation follows...       */

/*
Program: JIS.C
Version: 2.1
Date:    May 7, 1991
Author:  Ken R. Lunde, University of Wisconsin-Madison
EMAIL:   klunde@macc.wisc.edu or klunde@wiscmacc
MAIL:    610-G Eagle Heights, Madison, WI 53705 USA
Type:    A utility for converting the kanji code of Japanese textfiles.
Code:    ANSI C (portable)
 
    The most current version can be obtained through FTP at ucdavis.edu
    (128.120.2.1) in the pub/JIS/C directory, or by requesting a copy
    directly from me.
 
DESCRIPTION:

 1) Supports Shift-JIS, EUC, New-JIS, Old-JIS, and NEC-JIS for both input and
    output.

 2) Automatically detects input file's kanji code.  There is only one thing
    which will make this fail: a Shift-JIS file which only contains half-size
    katakana (remedy: insert a kanji anywhere in the file).

 3) The ability to convert half-size katakana in Shift-JIS and EUC into full-
    size equivalents.  Note that half-size katakana includes other symbols
    such as a Japanese period, Japanese comma, center dot, etc.

    The TOFULLSIZE flag determines whether to convert half-size katakana to
	their full-size counterparts.  If the output file is in New-JIS, Old-JIS,
    or NEC-JIS, ALL half-size katakana will be converted to their full-size
    counterparts regardless of the setting of TOFULLSIZE (necessary, since
    those 3 7-bit codes do not support half-size katakana).

 4) Supports conversion between the same code (i.e., EUC -> EUC, Shift-JIS ->
    Shift-JIS, etc.).  This is useful as a filter for converting half-size
    katakana to their full-size equivalents.

 5) If the input file does not contain any Japanese, then the input file's
    contents are simply echoed to the output file.  This adds reliability in
    case this program were used as a filter for incoming/outgoing mail.
*/


#include "jwp.h"

#include <stdio.h>


#define TOFULLSIZE		TRUE
#define UNGETBUFSIZ 	10

#define ReadChar()      ((stackp > 0) ? UngetBuf[--stackp] : f_charin())
#define WriteChar(x)    f_charout(x)
#define UnreadChar(x)   UngetBuf[stackp++] = (x)


#define NEW_KI      "$B"
#define NEW_KO      "(J"
#define OLD_KI      "$@"
#define OLD_KO      "(J"
#define NEC_KI      "K"
#define NEC_KO      "H"

#define NUL         0
#define NL          10
#define FF          12
#define CR          13
#define ESC         27
#define PERIOD      '.'

#define ISSJIS1(x)  (((x) >= 129 && (x) <= 159) || ((x) >= 224 && (x) <= 239))
#define ISSJIS2(x)  ((x) >= 64 && (x) <= 252)
#define ISHALFSIZE(x) ((x) >= 161 && (x) <= 223)
#define ISEUC(x)    ((x) >= 161 && (x) <= 254)
#define NOTEUC(x,y) (((x) >= 129 && (x) <= 159) && ((y) >= 64 && (y) <= 160))
#define ISMARU(x)   ((x) >= 202 && (x) <= 206)
#define ISNIGORI(x) (((x) >= 182 && (x) <= 196) || ((x) >= 202 && (x) <= 206))
#define ISODD(x)    ((x) % 2)
 
FILEFORMAT DetectCodeType (void);
BOOL SkipEscSeq (char temp);
void HalfToFull (int *p1, int *p2, FILEFORMAT incode);
void EnghtToSeven (int *p1, int *p2);
void SevenToEight (int *p1, int *p2);
void StraightEcho (void);
void ShiftToSeven (FILEFORMAT outcode, FILEFORMAT incode);
void ShiftToEuc (FILEFORMAT incode);
void EucToSeven (FILEFORMAT outcode, FILEFORMAT incode);
void EucToEuc (FILEFORMAT incode);
void ShiftToShift (FILEFORMAT incode);
void EucToShift (FILEFORMAT incode);
void SevenToShift (void);
void SevenToEuc (void);
void SevenToSeven (FILEFORMAT outcode);
void KanjiShiftIn (FILEFORMAT outcode);
void KanjiShiftOut (FILEFORMAT outcode);


/* The I/O Routines */

static int (* f_charin)(void);
static void (* f_charout)(int);

static char UngetBuf[UNGETBUFSIZ];
static int stackp;




void SetupIO (int (* in)(void), void (* out)(int))
{
	f_charin = in;
	f_charout = out;
	stackp = 0;
}


static void OutPuts (char *buf)
{
    for (; *buf; buf++) WriteChar(*buf);
}



/* Convert input file ==> EUC */

void FileImport (FILEFORMAT incode)
{
    switch (incode) {
		case FF_UNKNOWN:    StraightEcho(); break;

		case FF_OLDJIS:
        case FF_NEWJIS:
		case FF_NEC:        SevenToEuc(); break;

		case FF_EUC:        EucToEuc(incode); break;

		case FF_SJIS:       ShiftToEuc(incode); break;
	}
}


/* Convert internal EUC ==> Output format */

void FileExport (FILEFORMAT outcode)
{
    switch (outcode) {
        case FF_OLDJIS:
        case FF_NEWJIS:
        case FF_NEC:        EucToSeven(outcode,FF_EUC); break;

        case FF_UNKNOWN:
        case FF_EUC:        EucToEuc(FF_EUC); break;

		case FF_SJIS:       EucToShift(FF_EUC); break;
    }
}



static BOOL SkipEscSeq (char ch)
{
    if (ch == '$' || ch == '(') ReadChar();

    return (ch == 'K' || ch == '$');
}

 
static void KanjiShiftIn (FILEFORMAT outcode)
{
    switch (outcode) {
        case FF_NEWJIS: WriteChar(ESC); OutPuts(NEW_KI); break;
        case FF_OLDJIS: WriteChar(ESC); OutPuts(OLD_KI); break;
        case FF_NEC:    WriteChar(ESC); OutPuts(NEC_KI); break;
    }
}

 
static void KanjiShiftOut (FILEFORMAT outcode)
{
    switch (outcode) {
        case FF_NEWJIS: WriteChar(ESC); OutPuts(NEW_KO); break;
        case FF_OLDJIS: WriteChar(ESC); OutPuts(OLD_KO); break;
        case FF_NEC:    WriteChar(ESC); OutPuts(NEC_KO); break;
    }
}

 
static void EnghtToSeven (int *p1, int *p2)
{
    int temp;
 
    temp = *p2;

    if (*p2 >= 64 && *p2 <= 158)
        *p2 -= 31;
	else if (*p2 >= 159 && *p2 <= 252)
        *p2 -= 126;

    if (temp > 127 && temp <= 158) (*p2)--;

    if (*p1 >= 129 && *p1 <= 159) {
        if (temp >= 64 && temp <= 158)
            *p1 = (*p1 - 113) * 2 + 1;
        else if (temp >= 159 && temp <= 252)
            *p1 = (*p1 - 112) * 2;
    } else if (*p1 >= 224 && *p1 <= 239) {
        if (temp >= 64 && temp <= 158)
            *p1 = (*p1 - 177) * 2 + 1;
        else if (temp >= 159 && temp <= 252)
            *p1 = (*p1 - 176) * 2;
    }
}

 
static void SevenToEight (int *p1, int *p2)
{
    *p2 += (ISODD(*p1) ? 31 : 126);

    if (*p2 >= 127 && *p2 < 158) (*p2)++;

    if (*p1 >= 33 && *p1 <= 94) {
        if (ISODD(*p1))
            *p1 = (*p1 - 1) / 2 + 113;
        else /*if (!ISODD(*p1))*/
            *p1 = (*p1 / 2) + 112;
    } else if (*p1 >= 95 && *p1 <= 126) {
        if (ISODD(*p1))
            *p1 = (*p1 - 1) / 2 + 177;
        else /*if (!ISODD(*p1))*/
            *p1 = (*p1 / 2) + 176;
    }
}
 

static void StraightEcho (void)
{
  int p1;
 
  while ((p1 = ReadChar()) != EOF) WriteChar(p1);
}

 
static void ShiftToSeven (FILEFORMAT outcode, FILEFORMAT incode)
{
    BOOL ShiftedIn;
    int p1, p2;
 
    ShiftedIn = FALSE;

    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case NUL:
            case FF:    break;

            case CR:    break;

            case NL:    if (ShiftedIn) {
                            ShiftedIn = FALSE;
                            KanjiShiftOut(outcode);
                        }
                        WriteChar(CR);
                        WriteChar(NL);
                        break;

            default:    if (ISSJIS1(p1)) {
                            p2 = ReadChar();
                            if (ISSJIS2(p2)) {
                                EnghtToSeven(&p1, &p2);
                                if (!ShiftedIn) {
                                    ShiftedIn = TRUE;
                                    KanjiShiftIn(outcode);
                                }
                            }
                            WriteChar(p1);
                            WriteChar(p2);
                        } else if ISHALFSIZE(p1) {
                            HalfToFull(&p1, &p2, incode);
                            EnghtToSeven(&p1, &p2);
                            if (!ShiftedIn) {
                                ShiftedIn = TRUE;
                                KanjiShiftIn(outcode);
                            }
                            WriteChar(p1);
                            WriteChar(p2);
                        } else {
                            if (ShiftedIn) {
                                ShiftedIn = FALSE;
                                KanjiShiftOut(outcode);
                            }
                            WriteChar(p1);
                        }
                        break;
        }
    }


    if (ShiftedIn) KanjiShiftOut(outcode);
}


static void ShiftToEuc (FILEFORMAT incode)
{
    int p1, p2;
 
    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case CR:    break;

            case NL:    WriteChar(CR); WriteChar(NL); break;

            case NUL:
            case FF:    break;

            default:    if (ISSJIS1(p1)) {
                            p2 = ReadChar();
                            if (ISSJIS2(p2)) {
                                EnghtToSeven(&p1, &p2);
                                p1 += 128;
                                p2 += 128;
                            }
                            WriteChar(p1);
                            WriteChar(p2);
                        } else if ISHALFSIZE(p1) {
                            if (TOFULLSIZE) {
                                HalfToFull(&p1, &p2, incode);
                                EnghtToSeven(&p1, &p2);
                                p1 += 128;
                                p2 += 128;
                            } else {
								p2 = p1;
                                p1 = 142;
                            }
                            WriteChar(p1);
                            WriteChar(p2);
                        } else {
                            WriteChar(p1);
                        }
                        break;
        }
    }
}

 
static void EucToSeven (FILEFORMAT outcode, FILEFORMAT incode)
{
    BOOL ShiftedIn;
    int p1, p2;
 
    ShiftedIn = FALSE;

    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case NL:    if (ShiftedIn) {
                            ShiftedIn = FALSE;
                            KanjiShiftOut(outcode);
						}
                        WriteChar(CR);
                        WriteChar(NL);
						break;

            case FF:    break;

            default:    if (ISEUC(p1)) {
                            p2 = ReadChar();
                            if (ISEUC(p2)) {
                                p1 -= 128;
                                p2 -= 128;
                                if (!ShiftedIn) {
                                    ShiftedIn = TRUE;
                                    KanjiShiftIn(outcode);
                                }
                            }
                            WriteChar(p1);
                            WriteChar(p2);
                        } else if (p1 == 142) {
                            p2 = ReadChar();
                            if (ISHALFSIZE(p2)) {
                                p1 = p2;
                                HalfToFull(&p1, &p2, incode);
                                EnghtToSeven(&p1, &p2);
                                if (!ShiftedIn) {
                                    ShiftedIn = TRUE;
                                    KanjiShiftIn(outcode);
                                }
                            }
                            WriteChar(p1);
                            WriteChar(p2);
                        } else {
                            if (ShiftedIn) {
                                ShiftedIn = FALSE;
                                KanjiShiftOut(outcode);
                            }
                            WriteChar(p1);
                        }
                        break;
        }
    }

    if (ShiftedIn) KanjiShiftOut(outcode);
}

 
static void EucToShift (FILEFORMAT incode)
{
    int p1, p2;
 
    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case FF:    break;

            default:    if (ISEUC(p1)) {
                            p2 = ReadChar();
                            if (ISEUC(p2)) {
                                p1 -= 128;
                                p2 -= 128;
                                SevenToEight(&p1, &p2);
                            }
                            WriteChar(p1);
                            WriteChar(p2);
                        } else if (p1 == 142) {
                            p2 = ReadChar();
                            if (ISHALFSIZE(p2)) {
                                if (TOFULLSIZE) {
                                    p1 = p2;
                                    HalfToFull(&p1, &p2, incode);
                                    WriteChar(p1);
                                    WriteChar(p2);
                                } else {
									p1 = p2;
                                    WriteChar(p1);
                                }
                            } else {
                                WriteChar(p1);
                                WriteChar(p2);
                            }
                        } else {
                            WriteChar(p1);
                        }
                        break;
        }
    }
}
 

static void SevenToShift (void)
{
    BOOL ShiftedIn;
    int p1, p2;
 
    ShiftedIn = FALSE;

    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case ESC:   ShiftedIn = SkipEscSeq(ReadChar()); break;

            case FF:    break;

            default:    if (ShiftedIn) {
                            p2 = ReadChar();
                            SevenToEight(&p1, &p2);
                            WriteChar(p1);
                            WriteChar(p2);
                        } else {
                            WriteChar(p1);
                        }
                        break;
        }
    }
}

 
static void SevenToEuc (void)
{
    BOOL ShiftedIn;
    int p1, p2;
 
    ShiftedIn = FALSE;

    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case ESC:   ShiftedIn = SkipEscSeq(ReadChar()); break;

            case NL:    if (ShiftedIn) {
                            ShiftedIn = FALSE;
                        }
                        WriteChar(CR);
                        WriteChar(NL);
                        break;

            case FF:    break;

            default:    if (ShiftedIn) {
                            p2 = ReadChar();
                            p1 += 128;
                            p2 += 128;
                            WriteChar(p1);
                            WriteChar(p2);
                        } else {
                            WriteChar(p1);
                        }
                        break;
        }
    }
}
 

static void SevenToSeven (FILEFORMAT outcode)
{
    BOOL ShiftedIn;
    int p1, p2;
 
    ShiftedIn = FALSE;

    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case ESC:   ShiftedIn = SkipEscSeq(ReadChar());
                        if (ShiftedIn) KanjiShiftIn(outcode); else KanjiShiftOut(outcode);
                        break;

            case NL:    if (ShiftedIn) {
                            ShiftedIn = FALSE;
                            KanjiShiftOut(outcode);
                        }
                        WriteChar(CR);
                        WriteChar(NL);
                        break;

            case FF:    break;

            default:    if (ShiftedIn) {
                            p2 = ReadChar();
                            WriteChar(p1);
                            WriteChar(p2);
                        } else {
                            WriteChar(p1);
                        }
                        break;
        }
    }

    if (ShiftedIn) KanjiShiftOut(outcode);
}

 
static void EucToEuc (FILEFORMAT incode)
{
    int p1, p2;
 
    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case FF:    break;

            default:    if (ISEUC(p1)) {
                            p2 = ReadChar();
                            if (ISEUC(p2)) {
                                WriteChar(p1);
                                WriteChar(p2);
                            }
                        } else if (p1 == 142) {
                            p2 = ReadChar();
                            if (ISHALFSIZE(p2) && TOFULLSIZE) {
                                p1 = p2;
                                HalfToFull(&p1, &p2, incode);
                                EnghtToSeven(&p1, &p2);
                                p1 += 128;
                                p2 += 128;
                            }
                            WriteChar(p1);
                            WriteChar(p2);
                        } else {
                            WriteChar(p1);
                        }
                        break;
        }
    }
}

 
static void ShiftToShift (FILEFORMAT incode)
{
    int p1, p2;
 
    while ((p1 = ReadChar()) != EOF) {
        switch (p1) {
            case CR:    break;

            case NL:    WriteChar(CR); WriteChar(NL); break;

            case NUL:
            case FF:    break;

            default:    if (ISSJIS1(p1)) {
                            p2 = ReadChar();
                            if (ISSJIS2(p2)) {
                                WriteChar(p1);
                                WriteChar(p2);
                            }
                        } else if (ISHALFSIZE(p1) && TOFULLSIZE) {
                            HalfToFull(&p1, &p2, incode);
                            WriteChar(p1);
                            WriteChar(p2);
                        } else {
                            WriteChar(p1);
                        }
                        break;
        }
    }
}
 

FILEFORMAT DetectCodeType (void)
{
    int p1, p2, p3;
    FILEFORMAT incode;
 
    incode = FF_UNKNOWN;

    while ((p1 = ReadChar()) != EOF) {
        if (incode != FF_UNKNOWN && incode != FF_EUC) break;

        if (p1 == ESC) {
            p2 = ReadChar();
            if (p2 == '$') {
                p3 = ReadChar();

                if (p3 == 'B') incode = FF_NEWJIS;
                else if (p3 == '@') incode = FF_OLDJIS;

            } else if (p2 == 'K') {
                incode = FF_NEC;
            }
        } else if (p1 >= 129 && p1 <= 254) {
            p2 = ReadChar();

            if (NOTEUC(p1,p2)) incode = FF_SJIS;
            else if (ISEUC(p1) && ISEUC(p2)) incode = FF_EUC;
            else if (p1 == 142 && ISHALFSIZE(p2)) incode = FF_EUC;
        }
    }

    return (incode);
}

 
static void HalfToFull (int *one, int *two, FILEFORMAT incode)
{
    int junk;
    BOOL maru, nigori;
 
    maru = nigori = FALSE;

    if (incode == FF_SJIS) {
        *two = ReadChar();
        if (*two == 222) {
            if (ISNIGORI(*one) || *one == 179) nigori = TRUE; else UnreadChar(*two);
        } else if (*two == 223) {
            if (ISMARU(*one)) maru = TRUE; else UnreadChar(*two);
        } else {
            UnreadChar(*two);
        }
    } else if (incode == FF_EUC) {
        junk = ReadChar();
        if (junk == 142) {
            *two = ReadChar();
            if (*two == 222) {
                if (ISNIGORI(*one) || *one == 179) nigori = TRUE;
                else {
                    UnreadChar(*two);
                    UnreadChar(junk);
                }
            } else if (*two == 223) {
                if (ISMARU(*one)) maru = TRUE;
                else {
                    UnreadChar(*two);
                    UnreadChar(junk);
                }
            } else {
                UnreadChar(*two);
                UnreadChar(junk);
            }
        } else {
            UnreadChar(junk);
        }
    }

    switch (*one) {
        case 161:   *one = 129; *two = 66; break;
        case 162:   *one = 129; *two = 117; break;
        case 163:   *one = 129; *two = 118; break;
        case 164:   *one = 129; *two = 65; break;
        case 165:   *one = 129; *two = 69; break;
        case 166:   *one = 131; *two = 146; break;
        case 167:   *one = 131; *two = 64; break;
        case 168:   *one = 131; *two = 66; break;
        case 169:   *one = 131; *two = 68; break;
        case 170:   *one = 131; *two = 70; break;
        case 171:   *one = 131; *two = 72; break;
        case 172:   *one = 131; *two = 131; break;
        case 173:   *one = 131; *two = 133; break;
        case 174:   *one = 131; *two = 135; break;
        case 175:   *one = 131; *two = 98; break;
        case 176:   *one = 129; *two = 91; break;
        case 177:   *one = 131; *two = 65; break;
        case 178:   *one = 131; *two = 67; break;
        case 179:   *one = 131; *two = 69; break;
        case 180:   *one = 131; *two = 71; break;
        case 181:   *one = 131; *two = 73; break;
        case 182:   *one = 131; *two = 74; break;
        case 183:   *one = 131; *two = 76; break;
        case 184:   *one = 131; *two = 78; break;
        case 185:   *one = 131; *two = 80; break;
        case 186:   *one = 131; *two = 82; break;
        case 187:   *one = 131; *two = 84; break;
        case 188:   *one = 131; *two = 86; break;
        case 189:   *one = 131; *two = 88; break;
        case 190:   *one = 131; *two = 90; break;
        case 191:   *one = 131; *two = 92; break;
        case 192:   *one = 131; *two = 94; break;
        case 193:   *one = 131; *two = 96; break;
        case 194:   *one = 131; *two = 99; break;
        case 195:   *one = 131; *two = 101; break;
        case 196:   *one = 131; *two = 103; break;
        case 197:   *one = 131; *two = 105; break;
        case 198:   *one = 131; *two = 106; break;
        case 199:   *one = 131; *two = 107; break;
        case 200:   *one = 131; *two = 108; break;
		case 201:   *one = 131; *two = 109; break;
        case 202:   *one = 131; *two = 110; break;
        case 203:   *one = 131; *two = 113; break;
        case 204:   *one = 131; *two = 116; break;
        case 205:   *one = 131; *two = 119; break;
        case 206:   *one = 131; *two = 122; break;
        case 207:   *one = 131; *two = 125; break;
        case 208:   *one = 131; *two = 126; break;
        case 209:   *one = 131; *two = 128; break;
        case 210:   *one = 131; *two = 129; break;
        case 211:   *one = 131; *two = 130; break;
        case 212:   *one = 131; *two = 132; break;
        case 213:   *one = 131; *two = 134; break;
		case 214:   *one = 131; *two = 136; break;
        case 215:   *one = 131; *two = 137; break;
        case 216:   *one = 131; *two = 138; break;
        case 217:   *one = 131; *two = 139; break;
        case 218:   *one = 131; *two = 140; break;
        case 219:   *one = 131; *two = 141; break;
        case 220:   *one = 131; *two = 143; break;
        case 221:   *one = 131; *two = 147; break;
        case 222:   *one = 129; *two = 74; break;
        case 223:   *one = 129; *two = 75; break;
	}

    if (nigori) {
		if ((*two >= 74 && *two <= 103) || (*two >= 110 && *two <= 122))
            (*two)++;
        else if (*one == 131 && *two == 69)
            *two = 148;
    } else if (maru && *two >= 110 && *two <= 122)
        *two += 2;
}
